﻿using System;
using System.Collections.Generic;
// demo program for classification and prediction using adaptive boosting
// James McCaffrey

namespace AdaptiveBoosting
{
  class AdaptiveBoostingProgram
  {
    static void Main(string[] args)
    {
      try
      {
        Console.WriteLine("\nBegin adaptive boosting classification demo\n");
 
        string[] features = new string[] { "Opponent", "Field", "Spread", "Result" };  // dependent feature to predict is last listed

        string[][] values = new string[4][];
        values[0] = new string[] { "Atlanta", "Buffalo", "Chicago", "Detroit" };  // opponent
        values[1] = new string[] { "Home", "Away" };                          // field
        values[2] = new string[] { "Small ", "Medium", "Large " };            // point spread (note: some spaces added)
        values[3] = new string[] { "Lose", "Win" };                           // the dependent/predicted variable

        string[][] rawTrain = new string[10][];  // typically training data would be in a text file or SQL table
        rawTrain[0] = new string[] { "Detroit", "Home", "Large ", "Win" };
        rawTrain[1] = new string[] { "Detroit", "Away", "Medium", "Win" };
        rawTrain[2] = new string[] { "Buffalo", "Home", "Small ", "Win" };
        rawTrain[3] = new string[] { "Buffalo", "Home", "Medium", "Win" };
        rawTrain[4] = new string[] { "Atlanta", "Away", "Large ", "Win" };
        rawTrain[5] = new string[] { "Chicago", "Home", "Medium", "Win" };

        rawTrain[6] = new string[] { "Chicago", "Away", "Small ", "Lose" };
        rawTrain[7] = new string[] { "Chicago", "Home", "Small ", "Lose" };
        rawTrain[8] = new string[] { "Atlanta", "Away", "Medium", "Lose" };
        rawTrain[9] = new string[] { "Detroit", "Away", "Large ", "Lose" };

        Console.WriteLine("Raw (string) training data for team Seattle:\n");
        Console.WriteLine("Opponent Field  Spread   Result");
        Console.WriteLine("===============================");
        ShowMatrix(rawTrain);

        Console.WriteLine("\nConverting and storing training data.");
        int[][] train = RawTrainToInt(rawTrain, values);  // training tuples stored as ints in a matrix; last col (Y) as -1 or +1
        Console.WriteLine("Training data in int form:\n");
        ShowMatrix(train, true);

        Console.WriteLine("\nCreating weak categorical stump learners from training data.");
        List<Learner> learners = MakeLearners(values, train);
        Console.WriteLine("Completed. Weak learners are:\n");
        for (int i = 0; i < learners.Count; ++i)
          Console.WriteLine("[" + i + "] " + Description(learners[i], features, values));

        Console.WriteLine("\nInitializing list of best learner indexes.");
        List<int> bestLearners = new List<int>();  // indexes of best weak learner found for each iteration t in main algorithm loop
        
        Console.WriteLine("\nUsing adaptive boosting to find best learners and their alpha values.");
        MakeModel(train, values, learners, bestLearners);  // find best learners and their alphas (importance)

        Console.WriteLine("\nModel completed.");
        int numGood = bestLearners.Count;
        Console.WriteLine("Algorithm found " + numGood + " good learners and associated alpha values.");  // some weak learners may not make the cut

        Console.WriteLine("\nThe good learners and their alpha value are:");
        for (int i = 0; i < bestLearners.Count; ++i)
        {
          int lrn = bestLearners[i];
          Console.Write("[" + lrn + "] " + learners[lrn].alpha.ToString("F2") + "  ");
        }

        //Console.WriteLine("\nPredicting outcome when Opponent = Buffalo, Field = Away, Spread = Small\n");
        //int[] unknownTuple = new int[] { 1, 1, 0 };    // Buffalo, Away, Small

        Console.WriteLine("\nPredicting outcome when Opponent = Detroit, Field = Home, Spread = Small.\n");
        int[] unknownTuple = new int[] { 3, 0, 0 };    // Detroit, Home, Small

        int Y = Classify(unknownTuple, learners, bestLearners);
        Console.WriteLine("Predicted Y = " + Y + " => Seattle will " + YValueToString(Y, values));

        Console.WriteLine("\nEnd\n");
        Console.ReadLine();
      }
      catch (Exception ex)
      {
        Console.WriteLine(ex.Message);
        Console.ReadLine();
      }

    } // Main

    // data conversion routines below

    static int[][] RawTrainToInt(string[][] rawTrain, string[][] values)
    {
      // converts raw training data in string form to 0-based ints; last col to -1 or +1
      // calls helper ValueToInt
      int rows = rawTrain.Length;
      int cols = rawTrain[0].Length;
      int[][] result = new int[rows][];
      for (int i = 0; i < rows; ++i)
        result[i] = new int[cols];

      for (int i = 0; i < rawTrain.Length; ++i)
      {
        for (int j = 0; j < rawTrain[i].Length; ++j)  // all cols
        {
          int feature = j;
          int v = ValueToInt(rawTrain[i][j], feature, values);  // string to int
          result[i][j] = v;
        }
      }

      for (int i = 0; i < result.Length; ++i)  // convert values in last col from 0,1 to -1,+1
        if (result[i][cols - 1] == 0) result[i][cols - 1] = -1;

      return result;
    }  // RawTrainToInt

    static int ValueToInt(string value, int feature, string[][] values)
    {
      // convert a string value like "Detroit" to an int (3)
      int row = feature;
      for (int j = 0; j < values[row].Length; ++j)
        if (values[row][j] == value)
          return j;
      return -1; // error
    }

    static string IntToValue(int value, int feature, string[][] values)
    {
      // convert an int value like 2 to a string ("Chicago")
      return values[feature][value];
    }

    static string YValueToString(int Y, string[][] values)
    {
      // convert -1 or +1 to a string like "Lose" or "Win"
      if (Y == -1) return values[values.Length - 1][0];
      else if (Y == 1) return values[values.Length - 1][1];
      else return "error";
    }

    static int FeatureToInt(string feature, string[] features)
    {
      // convert a string feature like "Opponent" to an int (0)
      for (int i = 0; i < features.Length; ++i)
        if (features[i] == feature)
          return i;
      return -1;  // error
    }

    static string IntToFeature(int feature, string[] features)
    {
      // convert an int like 1 to a string ("Field")
      return features[feature];
    }

    // learner-making routines below. see class Learner too.

    static string Description(Learner learner, string[] features, string[][] values)
    {
      // return a friendly string description of a weak Learner
      string feature = IntToFeature(learner.feature, features);  // like "Opponent"
      string value = IntToValue(learner.value, learner.feature, values);  // like "Atlanta"
      string dependent = features[features.Length - 1];
      string predicted = "";
      if (learner.predicted == 1)
        predicted = values[values.Length - 1][1];
      else  // predicted == -1
        predicted = values[values.Length - 1][0];
  
      return "IF " + feature + " IS " + value + " THEN " + dependent + " IS " + predicted + "   (raw error = " + learner.error.ToString("F2") + ")";
    }
 
    static List<Learner> MakeLearners(string[][] values, int[][] train)
    {
      // scan through training set and try to make one weak learner for each possible feature value
      // do not create a learner for values with raw error rate of 0.50 or greater
      int numFeatures = values.Length;  // includes the dependent variable
      List<Learner> result = new List<Learner>();

      for (int f = 0; f < numFeatures - 1; ++f)  // each predictive feature (but not Y)
      {
        int numValues = values[f].Length;    // num values for curr feature
        for (int v = 0; v < numValues; ++v)  // each value of curr feature
        {
          int pred = MostLikely(f, v, train);  // most likely Y (-1,+1) for v; 0 if no most likely can be determined or error >= 0.50
          if (pred == 0) continue;  // 0 means that we could not make a prediction for this value, so we won't make a Learner for this feature value
          Learner lrn = new Learner(f, v, pred, -1.0, -1.0, -1.0);  // tmp error -1.0, tmp epsilon -1.0, tmp alpha = -1.0
          result.Add(lrn);
        }
      }
      // for each learner, scan training data and compute raw error rate
      for (int i = 0; i < result.Count; ++i)  // each learner
      {
        int nw = 0;  // number wrong
        int nr = 0;  // number relevant
        for (int k = 0; k < train.Length; ++k)  // each training tuple
        {
          if (IsApplicable(result[i], train[k]) == false) continue;  // the curr rule doesn't apply to curr tuple so skip
          ++nr;
          int act = train[k][train[k].Length - 1];  // actual -1 or +1 is value in last column of training matrix
          if (result[i].predicted != act) ++nw;  // wrong
        }
        if (nr == 0) result[i].error = 0.0;  // consider throWing an Exception here
        else result[i].error = (nw * 1.0) / nr;
      }

      return result;
    }

    static int MostLikely(int feature, int value, int[][] train)
    {
      // highest probability Y value for a particular feature value (like "field" is "Home")
      // return -1 or +1 based on highest prob,
      // return 0 if equal percentages (0.5 and 0.5) or no relevant training data
      // used to determine the predicted Y field in a Learner object in method MakeLearners
      int ctMinusOne = 0;  // relevant count of -1 values: number training tuples where value has Y = -1
      int ctPlusOne = 0;  // relevant count +1 values
      int totRelevant = 0;
      int cols = train[0].Length;

      for (int i = 0; i < train.Length; ++i)  // each training tuple
      {
        if (train[i][feature] != value) continue;  // this tuple not relevant because wrong value

        ++totRelevant;  // we know curr tuple is relevant

        if (train[i][cols - 1] == -1)   // the last column is the dependent variable (like sex)
          ++ctMinusOne;
        else
          ++ctPlusOne;
      }

      if (totRelevant == 0) return 0;    // there were no relevant training tuples for the feature value so cannot determine a most likely Y
      if (ctMinusOne > 0 && ctPlusOne == 0) return -1;  // there was at least one -1 but no +1 values so most likely Y is -1
      if (ctMinusOne == 0 && ctPlusOne > 0) return 1;   // see comment above

      double pctMinusOne = (ctMinusOne * 1.0) / totRelevant;  // percentage or probability that Y is -1
      double pctPlusOne = (ctPlusOne * 1.0) / totRelevant;     // percentage or probability that Y is +1

      if (VeryCLose(pctMinusOne, pctPlusOne) == true) return 0;  // both pcts are 0.5000 so no most likely Y

      if (pctMinusOne > pctPlusOne) return -1; else return 1;
    }

    // model-building routines below

    static void UpdateEpsilons(List<Learner> learners, int[][] train, double[] trainWeights)
    {
      // for each learner, update its epsilon by walking thru each training tuple and computing a weighted error sum
      for (int i = 0; i < learners.Count; ++i)
      {
        double ep = 0.0;
        for (int k = 0; k < train.Length; ++k)  // each training tuple
        {
          if (IsApplicable(learners[i], train[k]) == false) continue;  //  curr learner doesn't apply to curr tuple so move to next tuple
          int actual = train[k][train[k].Length - 1];  // actual -1 +1 is in last col of training matrix
          int pred = learners[i].predicted;  // predicted -1 +1
          if (actual != pred)  // incorrect prediction so . . 
            ep = ep + trainWeights[k];  // accumulate weighted error
        }
        learners[i].epsilon = ep;  // could be zero -- must deal with before computing alpha 
      }  // each learner
    }

    static bool IsApplicable(Learner learner, int[] tuple)
    {
      // is the learner/weak classifer applicable to the training tuple?
      int ti = learner.feature;  // index into the tuple
      if (tuple[ti] == learner.value) return true; else return false;
    }

    static int BestLearner(List<Learner> learners, out double epsilon, int[] usedLearners)
    {
      // find index of not-yet-used learner that has Smallest epsilon (weighted error)
      // update usedLearners hash table array

      // check to make sure there is at least one unused weak learner
      int x = Array.IndexOf(usedLearners, 0);  // find first unused learner
      if (x < 0) throw new Exception("No unused weak learners found by method BestLearner");

      double SmallestEpsilon = double.MaxValue;  // Smaller epsilon values mean Smaller error which means better learner
      int bestIndex = -1;
      for (int i = 0; i < learners.Count; ++i)
      {
        if (usedLearners[i] == 1) continue;  // curr weak learner has already been used

        if (learners[i].epsilon < SmallestEpsilon)
        {
          SmallestEpsilon = learners[i].epsilon;
          bestIndex = i;
        }
      }
 
      if (SmallestEpsilon < 0.0) throw new Exception("Encountered a negative epsilon in method BestLearner");
      if (bestIndex == -1) throw new Exception("Unable to find best learner index in method BestLearner");
      epsilon = SmallestEpsilon;    // assign out parameter value. could be 0.0
      usedLearners[bestIndex] = 1;  // flag the best weak learner as used
      return bestIndex;
    }

    static bool CLoseToZero(double x)
    {
      if (Math.Abs(x) < 0.00000001) return true;
      else return false;
    }

    static bool VeryCLose(double x, double y)
    {
      if (Math.Abs(x - y) < 0.00000001) return true;
      else return false;
    }

    static void UpdateTrainingWeights(double[] trainWeights, int[][] train, List<Learner> learners, double alphaT, int bestIndex)
    {
      // increase weights of misclassified (by curr best learner) training tuples, decrease weights of correctly classified tuples
      int cols = train[0].Length;  // number columns of training data matrix

      for (int i = 0; i < trainWeights.Length; ++i)  // each weight and each training tuple
      {
        if (IsApplicable(learners[bestIndex], train[i]) == false) continue;
        int actual = train[i][cols - 1];       // -1 or +1
        trainWeights[i] = trainWeights[i] * Math.Exp(-alphaT * actual * learners[bestIndex].predicted);  // tricky; depends on -1 +1 encoding for Y
      }

      double Z = 0.0;  // sum of all D weights to normalize them
      for (int i = 0; i < trainWeights.Length; ++i)  // each updated weight
        Z += trainWeights[i];

      for (int i = 0; i < trainWeights.Length; ++i)  // each updated weight
        trainWeights[i] = trainWeights[i] / Z;  // normalize (so all weights now sum to 1.0)
    }

    static void MakeModel(int[][] train, string[][] values, List<Learner> learners, List<int> bestLearners)
    {
      // compute (indexes of) best learners and their alphas
      // assumes all weak learners have been created
      // assumes bestLearners (index List) has been initialized to a empty List
      
      Console.WriteLine("\nInitializing training tuple weights ('D') array.");
      double[] trainWeights = new double[train.Length];  // training weights (D values) are local because they aren't needed to classify
      for (int i = 0; i < trainWeights.Length; ++i)      // one weight per tuple
        trainWeights[i] = 1.0 / trainWeights.Length;     // initially all weights equal

      //ShowVector(trainWeights);

      int[] usedLearners = new int[learners.Count];  // track which weak learners have been added to bestLearners List. index = which learner, value 0 = not used, 1 = used
      Console.WriteLine("Entering main algorithm loop.");
      bool done = false;
      int t = 0;  // not sure why research literature uses t for counter variable name here . . . 
      while (done == false)
      {
        //Console.WriteLine("================");
        //Console.WriteLine("t = " + t);
        UpdateEpsilons(learners, train, trainWeights);  // compute weighted error for each learner
        //for (int j = 0; j < learners.Count; ++j)
        //  Console.WriteLine(learners[j].ToString());

        double epsilonT;  // best (Smallest) weighted error at curr iteration t
        int bestIndex = BestLearner(learners, out epsilonT, usedLearners);  // find index of best learner (that hasn't yet been used) and its weighted error epsilonT
        //Console.WriteLine("best learner is " + bestIndex); Console.ReadLine();
        if (Math.Abs(0.5 - epsilonT) <= 0.00001) break;  // if best weighted error is (almost) 0.5 or greater then no use in continuing on
        bestLearners.Add(bestIndex);

        if (CLoseToZero(epsilonT) == true) epsilonT = 0.000001;  // avoid division by 0 when computing alphaT
        double alphaT = 0.5 * Math.Log((1.0 - epsilonT) / epsilonT);  // Smaller epsilon error makes Larger alpha values
        //Console.WriteLine("a = " + alphaT); Console.ReadLine();
        learners[bestIndex].alpha = alphaT;  // store the alpha

        UpdateTrainingWeights(trainWeights, train, learners, alphaT, bestIndex);  // update weights of training tuples that are relevant to best learner

        ++t;
        if (t == learners.Count) done = true;  // every weak learner has been used
        //Console.WriteLine("================");
      }
      bestLearners.Sort();  // sort this List of indexes to make the output easier to read. not absolutely necessary
    }

    static int Classify(int[] testTuple, List<Learner> learners, List<int> bestLearners)
    {
      // use adaptive boosting model (set of weak learners and their alpha (importtance) weights) to predict Y for a testTuple (in int form)
      double accVote = 0.0;  // accumulated votes of learners
      if (bestLearners.Count == 0) throw new Exception("No learners found in List of bestLearners");

      for (int t = 0; t < bestLearners.Count; ++t)
      {
        int idx = bestLearners[t];  // index of curr learner
        if (IsApplicable(learners[idx], testTuple) == false) continue;  // curr learner can't vote on the tuple

        //double vote = alphas[t] * learners[idx].predicted;
        double vote = learners[idx].predicted * learners[idx].alpha;
        accVote += vote;
        string sign = "";
        if (learners[idx].predicted == 1) sign = "+";
        //Console.WriteLine("Using learner " + idx + " with alpha = " + alphas[t].ToString("F2") + " prediction is " + sign + learners[idx].predicted + " so vote = " + vote.ToString("F2") );
        Console.WriteLine("Using learner " + idx + " with alpha = " + learners[idx].alpha.ToString("F2") + " prediction is " + sign + learners[idx].predicted + " so vote = " + vote.ToString("F2"));
      }

      Console.WriteLine("\nFinal accumulated vote is " + accVote.ToString("F2"));
      if (accVote > 0.0) return 1;
      else if (accVote < 0.0) return -1;
      else return 0;
    }

    // display routines below

    static void ShowVector(string[] vector)
    {
      for (int i = 0; i < vector.Length; ++i)
        Console.Write(vector[i] + " ");
      Console.WriteLine("");
    }

    static void ShowVector(int[] vector)
    {
      for (int i = 0; i < vector.Length; ++i)
        Console.Write(vector[i] + " ");
      Console.WriteLine("");
    }

    static void ShowVector(double[] vector)
    {
      for (int i = 0; i < vector.Length; ++i)
        Console.Write(vector[i].ToString("F4") + " ");
      Console.WriteLine("");
    }

    static void ShowMatrix(string[][] matrix)
    {
      for (int i = 0; i < matrix.Length; ++i)
      {
        for (int j = 0; j < matrix[i].Length; ++j)
        {
          Console.Write(matrix[i][j] + "   ");
        }
        Console.WriteLine("");
      }
    }

    static void ShowMatrix(int[][] matrix, bool isTrain)
    {
      for (int i = 0; i < matrix.Length; ++i)
      {
        for (int j = 0; j < matrix[i].Length - 1; ++j)
        {
          Console.Write(matrix[i][j] + " ");
        }
        if (isTrain == true)
        {
          Console.Write(" ->  ");
          if (matrix[i][matrix[i].Length - 1] == -1)
            Console.Write("-1");
          else
            Console.Write("+1");
        }
        Console.WriteLine("");
      }
    }

  } // class Program

  public class Learner  // weak learner
  {
    public int feature;  // which feature: Opponent, Field, Spread
    public int value;    // value: Atlanta, . . Home, . . Small, . .
    public int predicted;  // -1 (Lose) or +1 (Win)
    public double error;  // actual error rate is percentage of tuples incorrectly classified
    public double epsilon;  // weighted error (because different training tuples have different weights)
    public double alpha;  // importance weight 

    public Learner(int feature, int value, int predicted, double error, double epsilon, double alpha)
    {
      this.feature = feature;  // ex: "Opponent"
      this.value = value;      // ex: "Chicago"
      this.predicted = predicted;  // -1 or +1 (Lose, Win)
      this.error = error;    // ex: 0.45 (cannot be greater than 0.50)
      this.epsilon = epsilon;      // ex: 0.15 but not 0.65 (i.e., not greater than 0.50)
      this.alpha = alpha;  // ex: 2.36
    }

    public override string ToString()
    {
      string s = "";
      s += "feature = " + feature;
      s += " value = " + value;
      s += " predicted = " + ((predicted == 1.0) ? "+" : "") + predicted;
      s += " error = " + error.ToString("F4");
      s += " epsilon = " + epsilon.ToString("F4");
      s += " alpha = " + alpha.ToString("F4");
      return s;
    }

  } // Learner

} // ns

